module DataSpec where
	
import Auxiliary
import Ratio
import Char
import Debug.Trace

data DataType    = EnumInt String Int Int | EnumString String [String] deriving (Show, Eq)
data FunctionDef = FunctionDef String [([String], String)] deriving (Show, Eq)
data Type        = TypeName String | TypeRange Int Int | TypeRangeExpressions Expression Expression | NoType deriving (Show, Eq)

type TypeValues  = (Type, [String])

type DataSpec      = ([DataType], [TypeValues], [FunctionDef])

builtInTypes = [EnumString "Bool" ["T", "F"]]
builtInFunctions = []

allValues :: Type -> [DataType] -> [String]
allValues (TypeRange from to) _ = [show i | i <- [from..to]]
allValues (TypeName "Bool") _   = ["T", "F"]
allValues (TypeName "Queue") _  = ["empty", "bla"]
allValues (TypeName typeName) []     = error("Error: Undefined type " ++ typeName ++ " used.")
allValues (TypeName typeName) ((EnumInt name from to):xs) | name == typeName = [show i | i <- [from..to]]
                                               | otherwise        = allValues (TypeName typeName) xs
allValues (TypeName typeName) ((EnumString name values):xs) | name == typeName = values
                                                 | otherwise        = allValues (TypeName typeName) xs

getValues :: DataSpec -> Type -> [String]
getValues _ (TypeRange from to)                      = [show i | i <- [from..to]]
getValues (_,typeValues,_) (TypeName name)  | length valuesSet == 1 = valuesSet!!0 
                                            | otherwise             = error("Error: Values for type " ++ name ++ ": " ++ show valuesSet ++ show typeValues)
  where
    valuesSet = [values | (typ,values) <- typeValues, typ == TypeName name]

getTypeName :: DataType -> String
getTypeName (EnumInt name _ _)  = name
getTypeName (EnumString name _) = name

getFunctionName :: FunctionDef -> String
getFunctionName (FunctionDef name mappings) = name

printDataSpec :: DataSpec -> String
printDataSpec ([], _, []) = ""
printDataSpec ([], _, [FunctionDef name list])      = printFunction name list
printDataSpec ([],values, ((FunctionDef name list):fs)) = printFunction name list ++ "\n" ++ printDataSpec ([], values, fs)
printDataSpec ([typ], _, [])                        = printDataType typ
printDataSpec ([typ], values, fs)                        = printDataType typ ++ "\n" ++ printDataSpec ([], values, fs)
printDataSpec ((typ:types), values, fs)                  = printDataType typ ++ "\n" ++ printDataSpec (types, values, fs)

printFunction name list = "    function " ++ name ++ " = " ++ commaList (map printFunctionMapping list)

printFunctionMapping :: ([String], String) -> String
printFunctionMapping (from,to) = (infixString from " * ") ++ " -> " ++ to

printDataType (EnumInt name from to) = "    type " ++ name ++ " = {" ++ show from ++ ".." ++ show to ++ "}"
printDataType (EnumString name pars) = "    type " ++ name ++ " = " ++ commaList pars

stringToInt :: String -> Int
stringToInt s | isInteger s = parseInteger s
              | otherwise   = error("Error: Could not parse " ++ s ++ " as an integer. Maybe you made a typing mistake or forgot to define a constant?")

evaluateFunction :: String -> [String] -> [FunctionDef] -> String
evaluateFunction "ifthenelse" [i,t,e] _ = if (i == "T") then t else if (i == "F") then e else error("Error: cannot evaluate '" ++ i ++ "'") 
evaluateFunction "eq" [x,y] _     | isNumber x && isNumber y = if (getFraction x == getFraction y) then "T" else "F"
                                  | otherwise       = if x == y then "T" else "F"
evaluateFunction "eq" (x:y:xs) functions = if (x == y && (evaluateFunction "eq" (y:xs)) functions == "T") then "T" else "F"
evaluateFunction "not" ["F"] _  = "T"
evaluateFunction "not" ["T"] _   = "F"
evaluateFunction "concat" [s1, s2] _ = s1 ++ s2
evaluateFunction "and" arguments _  = if (elem "F" arguments) then "F" else "T"
evaluateFunction "or" arguments _   = if (elem "T" arguments) then "T" else "F"
evaluateFunction "gt" [i, j] _      = if (getFraction i > getFraction j) then "T" else "F"
evaluateFunction "geq" [i, j] _     = if (getFraction i >= getFraction j) then "T" else "F"
evaluateFunction "leq" [i, j] _     = if (getFraction i <= getFraction j) then "T" else "F"
evaluateFunction "lt" [i, j] _      = if (getFraction i < getFraction j) then "T" else "F"
evaluateFunction "min" [i,j] _      = writeFraction (min (getFraction i) (getFraction j))
evaluateFunction "max" [i,j] _      = writeFraction (max (getFraction i) (getFraction j))
evaluateFunction "plus" [i,j] _     = writeFraction (getFraction i + getFraction j)
evaluateFunction "power" [i,j] _    | isInteger i && isInteger j = show ((parseInteger i) ^ (parseInteger j))
                                    | otherwise                  = error("Error: Cannot compute powers of non-integers.")
evaluateFunction "minus" [i,j] _    = writeFraction (getFraction i - getFraction j)
evaluateFunction "multiply" [x,y] _ = writeFraction ((getFraction x) * (getFraction y))
evaluateFunction "divide" [x,y] _   = writeFraction ((getFraction x) / (getFraction y))
evaluateFunction "mod" [i,j] _      | isInteger i && isInteger j = show (mod (parseInteger i) (parseInteger j))
                                    | otherwise                  = error("Error: Cannot do modulo computation with non-integers.")
evaluateFunction "addToQueue" [q,m] _  | q == "empty" = m
                                       | otherwise    = q ++ ";" ++ m
evaluateFunction "head"       [q] _      = takeWhile (/= ';') q
evaluateFunction "tail"       [q] _      = if elem ';' q then drop 1 (dropWhile (/= ';') q) else "empty"
evaluateFunction "size"       [q] _      = if q == "empty" then "0" else show (length [x | x <- q, x == ';'] + 1)

evaluateFunction functionName arguments [] = error("Error: Undefined function " ++ functionName ++ show arguments ++ " used.")
evaluateFunction functionName arguments ((FunctionDef name mapping):xs) | name == functionName = useMapping name mapping arguments
                                                                        | otherwise = evaluateFunction functionName arguments xs 

useMapping :: String -> [([String], String)] -> [String] -> String
useMapping fName [] arguments = error("Error: Function " ++ fName ++ " applied to constants " ++ show arguments ++ ", without having defined the result")
useMapping fName ((from, to):xs) arguments | from == arguments = to
                                     | otherwise         = useMapping fName xs arguments

allTypes :: [DataType] -> [Type]
allTypes [] = []
allTypes ((EnumInt name from to):xs)   = (TypeName name):(allTypes xs)
allTypes ((EnumString name values):xs) = (TypeName name):(allTypes xs)

-- This function checks whether a string is an element of one of the datatypes.
stringIsConstant :: DataSpec -> String -> Bool
stringIsConstant dataspec constant = isNumber constant || elem constant (concat [getValues dataspec t | t <- allTypes (fst3 dataspec)])

isInteger :: String -> Bool
isInteger i   = isInt
  where
    nr        = takeWhile isDigit (if (head i == '-') then tail i else i)  
    isInt     = (length nr > 0) && (nr == i || ('-':nr) == i || nr ++ ".0" == i || ('-':nr) ++ ".0" == i || 
                                    nr ++ "/1" == i || ('-':nr)++"/1" == i || nr ++ ".0/1" == i || ('-':nr) ++ ".0/1" == i)

parseInteger :: String -> Int
parseInteger i = read (takeWhile (\x -> x /= '.' && x /= '/') i)

isReal :: String -> Bool
isReal i = correct
  where
    beforePoint = takeWhile isDigit (if (head i == '-') then tail i else i)  
    afterPoint  = takeWhile isDigit (drop 1 (dropWhile (/= '.') i))
    correct     = length beforePoint > 0 && length afterPoint > 0 && (beforePoint ++ "." ++ afterPoint == i ||
                                             "-" ++ beforePoint ++ "." ++ afterPoint == i)

parseReal :: String -> Ratio Int
parseReal i = fraction
  where
    beforePoint = takeWhile (/= '.') i  
    afterPoint  = drop 1 (dropWhile (/= '.') i)
    number      = read (beforePoint ++ afterPoint)
    fraction    = number % 10^(length afterPoint)

isFraction :: String -> Bool
isFraction i = correct
  where
    beforeDash = takeWhile isDigit (if (head i == '-') then tail i else i)  
    afterDash  = takeWhile isDigit (drop 1 (dropWhile (/= '/') i))
    correct    = length beforeDash > 0 && length afterDash > 0 && (beforeDash ++ "/" ++ afterDash == i ||
	                                             "-" ++ beforeDash ++ "/" ++ afterDash == i)

parseFraction :: String -> Ratio Int
parseFraction i = (read beforeDash) % (read afterDash)
  where
    beforeDash = takeWhile (/= '/') i  
    afterDash  = drop 1 (dropWhile (/= '/') i)

writeFraction :: Ratio Int -> String
writeFraction i | denominator i == 1 = show (numerator i)
	            | otherwise          = show (numerator i) ++ "/" ++ show (denominator i)

getFraction :: String -> Ratio Int
getFraction frac | frac == "1"     = 1 % 1
                 | frac == "2"     = 2 % 1
                 | frac == "3"     = 3 % 1
                 | frac == "4"     = 4 % 1
                 | frac == "5"     = 5 % 1
                 | frac == "6"     = 6 % 1
                 | frac == "7"     = 7 % 1
                 | frac == "8"     = 8 % 1
                 | frac == "9"     = 9 % 1
                 | isInteger  frac = parseInteger  frac % 1
                 | isReal     frac = parseReal     frac
                 | isFraction frac = parseFraction frac
                 | otherwise       = error("Error: Cannot parse " ++ frac ++ " as a number.")

isNumber :: String -> Bool
isNumber i = i == "1" || isInteger i || isReal i || isFraction i