-----------------------------------------------
-- This module deals with detecting whether  --
-- parameters are changed or used in an LPPE --
-----------------------------------------------

module Usage where

import LPPE
import Expressions
import Auxiliary

-- When using these types, [(i,j)] means that parameter j is 
-- changed / used / directly used in summand i.
type Changed      = [(Int, Int)]
type Used         = [(Int, Int)]
type DirectlyUsed = [(Int, Int)]

----------------------------------------------------------------
-- Functions for checking whether or not a parameters is used --
----------------------------------------------------------------

-- This function checks whether a parameter is directly used
-- in a summand, meaning that it is used in either the action
-- parameters, the condition, or the probabilistic choices.
isDirectlyUsedInSummand :: PSummand -> Variable -> Bool
isDirectlyUsedInSummand (params, c, a, aps, probChoices, g) name = directlyUsed
  where
    overwrittenBySum  = elem name (map fst params)
    usedInActionPars  = variableInExpressions name aps
    usedInCondition   = variableInExpression name c
    overwrittenByProb = elem name (map fst3 probChoices)
    usedInProb        = variableInExpressions name (map thd3 probChoices)
    directlyUsed      = not overwrittenBySum && (usedInActionPars || usedInCondition || 
                                                 (not overwrittenByProb && usedInProb))

-- This function checks whether a parameter is used either directly,
-- or in the next state expression of another parameter.
--
-- The boolean parameter indicates whether or not a parameter is considered
-- used in a summand when it is only used to update itself.
isUsedInSummand :: PSummand -> Int -> Variable -> Bool -> Bool
isUsedInSummand (params, c, a, aps, probChoices, g) i name self = used
  where
    overwrittenBySum  = elem name (map fst params)
    usedInActionPars  = variableInExpressions name aps
    usedInCondition   = variableInExpression name c
    overwrittenByProb = elem name (map fst3 probChoices)
    usedInProb        = variableInExpressions name (map thd3 probChoices)
    usedInNextState   = variableInExpressions name ([g!!j | j <- [0..i-1]] ++ [g!!j | j <- [i+1..length(g)-1]])
                         || (self && variableInExpression name (g!!i) && not((g!!i) == Variable name))
    used              = not overwrittenBySum && (usedInActionPars || usedInCondition ||  
                                                (not overwrittenByProb && (usedInProb || usedInNextState)))

-- This function checks whether a parameter is changed in a summand.
-- This is the case when either the next state is an expression different
-- from just the parameter name itself, or is the parameter is used in a 
-- nondeterministic or probabilistic sum.
isChangedInSummand :: LPPE -> Int -> Int -> Bool
isChangedInSummand lppe summandNr parNr = nextState /= Variable parameter 
                                          || elem parameter (map fst params) 
                                          || elem parameter (map fst3 probChoices)
  where
    summand   = getPSummand lppe summandNr
    nextState = getNextState summand parNr
    parameter = fst ((getLPPEPars lppe)!!parNr)
    (params, c, a, aps, probChoices, g) = summand

-- Provides a list of elements of the form (i,j),
-- indicating that summand i changes parameter j.
-- For every pair (i,j) that is NOT in this list,
-- summand i does not change parameter j.
--
-- Note that the result of this function is an overapproximation
-- of the variables that are actually changed, as heuristics
-- are used to detect whether or not a variable will be changed.
-- It could be the case the a pair (i,j) is present in the
-- result, even though summand i will in fact never actually
-- change parameter j.
getChanged :: LPPE -> Changed
getChanged lppe = [(summandNr, parNr) | summandNr <- getPSummandNrs lppe, 
                                        parNr <- [0..length (getLPPEPars lppe) - 1],
                                        isChangedInSummand lppe summandNr parNr]

-- Provides a list of elements of the form (i,j),
-- indicating that summand i uses parameter j.
-- For every pair (i,j) that is NOT in this list,
-- summand i does not uses parameter j.
--
-- Note that the result of this function is an overapproximation
-- of the variables that are actually used, as heuristics
-- are used to detect whether or not a variable will be used.
-- It could be the case the a pair (i,j) is present in the
-- result, even though summand i will in fact never actually
-- use parameter j.
getUsed :: LPPE -> Used
getUsed lppe = [(summandNr, parNr) | summandNr <- getPSummandNrs lppe, 
                                     parNr <- getUsedInSummand lppe summandNr
               ]

-- Provides a list of elements of the form (i,j),
-- indicating that summand i directly uses parameter j.
-- For every pair (i,j) that is NOT in this list,
-- summand i does not directly use parameter j.
--
-- Note that the result of this function is an overapproximation
-- of the variables that are actually directly used, as heuristics
-- are used to detect whether or not a variable will be directly used.
-- It could be the case the a pair (i,j) is present in the
-- result, even though summand i will in fact never actually
-- use parameter j directly.
getDirectlyUsed :: LPPE -> DirectlyUsed
getDirectlyUsed lppe = [(summandNr, parNr) | summandNr <- getPSummandNrs lppe, 
                                             parNr <- getDirectlyUsedInSummand lppe summandNr]

getDirectlyUsedInSummand :: LPPE -> Int -> [Int]
getDirectlyUsedInSummand (LPPE name pars summands) i = getDirectlyUsedInSummand2 (summands!!i) (zip [0..length(pars) - 1] (map fst pars))

getDirectlyUsedInSummand2 :: PSummand -> [(Int, Variable)] -> [Int]
getDirectlyUsedInSummand2 summand [] = []
getDirectlyUsedInSummand2 summand ((i,name):is) | isDirectlyUsedInSummand summand name = i:(getDirectlyUsedInSummand2 summand is)
                                                | otherwise                           = getDirectlyUsedInSummand2 summand is

getChangedInNextStateNums :: PSummand -> [(Int, Variable)] -> [Int]
getChangedInNextStateNums summand [] = []
getChangedInNextStateNums summand ((i,name):is) | isChangedInNextState summand i name = i:(getChangedInNextStateNums summand is)
                                                | otherwise                           = getChangedInNextStateNums summand is

isChangedInNextState :: PSummand -> Int -> Variable -> Bool
isChangedInNextState (params, c, a, aps, probChoices, g) i name = not((g!!i) == Variable name) || (elem name (map fst params)) || (elem name (map fst3 probChoices))

getUsedInSummand :: LPPE -> Int -> [Int]
getUsedInSummand (LPPE name pars summands) i = getUsedInSummand2 (summands!!i) (zip [0..length(pars) - 1] (map fst pars))

getUsedInSummand2 :: PSummand -> [(Int, Variable)] -> [Int]
getUsedInSummand2 summand [] = []
getUsedInSummand2 summand ((i,name):is) | isUsedInSummand summand i name True = i:(getUsedInSummand2 summand is)
                                                | otherwise                           = getUsedInSummand2 summand is